@@ -11,6 +11,14 @@ import re
 INDENT = '    '
 
 
+class Error(Exception):
+    pass
+
+
+class ParseError(Error):
+    pass
+
+
 def bump_child_depth(obj, depth):
     children = getattr(obj, 'children', [])
     for child in children:
@@ -67,7 +75,7 @@ class Conf(object):
         for x in self.children:
             if name and isinstance(x, Key) and x.name == name:
                 filtered.append(x)
-            elif isinstance(x, Container) and x.__class__.__name__ == btype\
+            elif isinstance(x, Container) and x.__class__.__name__ == btype \
                     and x.value == name:
                 filtered.append(x)
             elif not name and btype and x.__class__.__name__ == btype:
@@ -164,7 +172,7 @@ class Container(object):
         for x in self.children:
             if name and isinstance(x, Key) and x.name == name:
                 filtered.append(x)
-            elif isinstance(x, Container) and x.__class__.__name__ == btype\
+            elif isinstance(x, Container) and x.__class__.__name__ == btype \
                     and x.value == name:
                 filtered.append(x)
             elif not name and btype and x.__class__.__name__ == btype:
@@ -309,9 +317,9 @@ class LimitExcept(Container):
 class Types(Container):
     """Container for MIME type mapping."""
 
-    def __init__(self, value, *args):
+    def __init__(self, *args):
         """Initialize."""
-        super(Types, self).__init__(value, *args)
+        super(Types, self).__init__('', *args)
         self.name = 'types'
 
 
@@ -436,49 +444,56 @@ def loads(data, conf=True):
             index += m.end()
             continue
 
-        m = re.compile(r'^\s*location\s*(.*?)\s*{', re.S).search(data[index:])
+        m = re.compile(r'^\s*location\s*([^;]*?)\s*{', re.S).search(data[index:])
         if m:
             l = Location(m.group(1))
             lopen.insert(0, l)
             index += m.end()
             continue
 
-        m = re.compile(r'^\s*if\s*(.*?)\s*{', re.S).search(data[index:])
+        m = re.compile(r'^\s*if\s*([^;]*?)\s*{', re.S).search(data[index:])
         if m:
             ifs = If(m.group(1))
             lopen.insert(0, ifs)
             index += m.end()
             continue
 
-        m = re.compile(r'^\s*upstream\s*(.*?)\s*{', re.S).search(data[index:])
+        m = re.compile(r'^\s*upstream\s*([^;]*?)\s*{', re.S).search(data[index:])
         if m:
             u = Upstream(m.group(1))
             lopen.insert(0, u)
             index += m.end()
             continue
 
-        m = re.compile(r'^\s*geo\s*(.*?)\s*{', re.S).search(data[index:])
+        m = re.compile(r'^\s*geo\s*([^;]*?)\s*{', re.S).search(data[index:])
         if m:
             g = Geo(m.group(1))
             lopen.insert(0, g)
             index += m.end()
             continue
 
-        m = re.compile(r'^\s*map\s*(.*?)\s*{', re.S).search(data[index:])
+        m = re.compile(r'^\s*map\s*([^;]*?)\s*{', re.S).search(data[index:])
         if m:
             g = Map(m.group(1))
             lopen.insert(0, g)
             index += m.end()
             continue
 
-        m = re.compile(r'^\s*limit_except\s*(.*?)\s*{', re.S).search(data[index:])
+        m = re.compile(r'^\s*limit_except\s*([^;]*?)\s*{', re.S).search(data[index:])
         if m:
             l = LimitExcept(m.group(1))
             lopen.insert(0, l)
             index += m.end()
             continue
 
-        m = re.compile(r'^(\s*)#\s*(.*?)\n', re.S).search(data[index:])
+        m = re.compile(r'^\s*types\s*{', re.S).search(data[index:])
+        if m:
+            l = Types()
+            lopen.insert(0, l)
+            index += m.end()
+            continue
+
+        m = re.compile(r'^(\s*)#[ \r\t\f]*(.*?)\n', re.S).search(data[index:])
         if m:
             c = Comment(m.group(2), inline='\n' not in m.group(1))
             if lopen and isinstance(lopen[0], Container):
@@ -500,13 +515,19 @@ def loads(data, conf=True):
             index += m.end()
             continue
 
-        s1 = r'("[^"]+"|\'[^\']+\'|[^\s;]+)'
-        s2 = r'("[^"]*"|\'[^\']*\'|[^\s;]*)'
-        s3 = r'(\s*[^;]*?)'
-        key = r'^\s*{}\s*{}{};'.format(s1, s2, s3)
-        m = re.compile(key, re.S).search(data[index:])
+        if ";" not in data[index:] and index+1 != len(data):
+            # If there is still something to parse, expect ';' otherwise
+            # the Key regexp can get stuck due to regexp catastrophic backtracking
+            raise ParseError("Config syntax, missing ';' at index: {}".format(index))
+
+        double = r'\s*"[^"]*"'
+        single = r'\s*\'[^\']*\''
+        normal = r'\s*[^;\s]*'
+        s1 = r'{}|{}|{}'.format(double, single, normal)
+        s = r'^\s*({})\s*((?:{})+);'.format(s1, s1)
+        m = re.compile(s, re.S).search(data[index:])
         if m:
-            k = Key(m.group(1), m.group(2) + m.group(3))
+            k = Key(m.group(1), m.group(2))
             if lopen and isinstance(lopen[0], (Container, Server)):
                 lopen[0].add(k)
             else:
